"""
This is just my version of the Textacy quote attributor. Will remove this if/when it gets implemented in Textacy!
"""

from . import constants
from spacy.tokens import Doc, Token, Span
from spacy.symbols import VERB, PUNCT
from operator import attrgetter
import regex as re
from typing import Literal, Iterable

def direct_quotations(doc: Doc):
    """
    
    """
    qtoks = [tok for tok in doc if tok.is_quote or (re.match(r"\n", tok.text))]
    qtok_idx_pairs = [(-1, -1)]
    for n, q in enumerate(qtoks):
        if (
            not bool(q.whitespace_)
            and q.i not in [q_[1] for q_ in qtok_idx_pairs]
            and q.i > qtok_idx_pairs[-1][1]
        ):
            for q_ in qtoks[n + 1 :]:
                if (ord(q.text), ord(q_.text)) in constants.QUOTATION_MARK_PAIRS:
                    qtok_idx_pairs.append((q.i, q_.i))
                    break
    qtok_idx_pairs = qtok_idx_pairs[1:]

    def filter_quote_tokens(tok):
        return any(qts_idx <= tok.i <= qte_idx for qts_idx, qte_idx in qtok_idx_pairs)

    for qtok_start_idx, qtok_end_idx in qtok_idx_pairs:
        content = doc[qtok_start_idx:qtok_end_idx]
        cue = None
        speaker = None

        if (
            len(content.text.split()) < constants.MIN_QUOTE_LENGTH
            # filter out titles of books and such, if possible
            or all(tok.is_title for tok in content if not (tok.is_punct or tok.is_stop))
        ):
            continue

        for window_sents in [
            windower(content, "overlap"),
            windower(content, "linebreaks"),
        ]:
            # get candidate cue verbs in window
            cue_candidates = [
                tok
                for sent in window_sents
                for tok in sent
                if tok.pos == VERB
                and tok.lemma_ in constants._reporting_verbs
                and not filter_quote_tokens(tok)
            ]
            cue_candidates = sorted(
                cue_candidates,
                key=lambda cc: min(
                    abs(cc.i - qtok_start_idx), abs(cc.i - qtok_end_idx)
                ),
            )
            for cue_cand in cue_candidates:
                if cue is not None:
                    break
                speaker_cands = [
                    speaker_cand
                    for speaker_cand in cue_cand.children
                    if speaker_cand.pos != PUNCT
                    and not filter_quote_tokens(speaker_cand)
                    and (
                        (speaker_cand.i >= qtok_end_idx)
                        or (speaker_cand.i <= qtok_start_idx)
                    )
                ]
                for speaker_cand in speaker_cands:
                    if speaker_cand.dep in constants._ACTIVE_SUBJ_DEPS:
                        cue = expand_verb(cue_cand)
                        speaker = expand_noun(speaker_cand)
                        break
            if content and cue and speaker:
                yield constants.DQTriple(
                    speaker=sorted(speaker, key=attrgetter("i")),
                    cue=sorted(cue, key=attrgetter("i")),
                    content=doc[qtok_start_idx : qtok_end_idx + 1],
                )
                break


def expand_noun(tok: Token) -> list[Token]:
    """Expand a noun token to include all associated conjunct and compound nouns."""
    tok_and_conjuncts = [tok] + list(tok.conjuncts)
    compounds = [
        child
        for tc in tok_and_conjuncts
        for child in tc.children
        if child.dep_ == "compound"
    ]
    return tok_and_conjuncts + compounds


def expand_verb(tok: Token) -> list[Token]:
    """Expand a verb token to include all associated auxiliary and negation tokens."""
    verb_modifiers = [
        child for child in tok.children if child.dep in constants._VERB_MODIFIER_DEPS
    ]
    return [tok] + verb_modifiers


def windower(quote: Span, method: Literal["overlap", "linebreaks"]) -> Iterable[Span]:
    """
    Finds the range of sentences in which to look for quote attribution.

    3 ways:
    - "overlap": any sentences that overlap with the quote span
    - "linebreaks": overlap sentences +/- one sentence, without crossing linebreaks after the quote
    - None: overlap sentences +/- one sentence,

    Input:
        quote (Span) - quote to be attributed
        method (str) - how the sentence range will be determined

    Output:
        sents (list) - list of sentences
    """
    if method == "overlap":
        return [
            sent
            for sent in quote.doc.sents
            if (sent.start < quote.start < sent.end)
            or (sent.start < quote.end < sent.end)
        ]
    else:
        sent_indexes = [
            n
            for n, s in enumerate(quote.doc.sents)
            if (s.start <= quote.start <= s.end) or (s.start <= quote.end <= s.end)
        ]

        i_sent = sent_indexes[0] - 1 if sent_indexes[0] > 0 else 0
        j_sent = sent_indexes[-1] + 2
        sents = list(quote.doc.sents)[i_sent:j_sent]
        if method == "linebreaks":
            linebreaks = (
                [0]
                + [tok.i for tok in quote.doc if re.match(r"\n", tok.text)]
                + [quote.doc[-1].i]
            )
            linebreak_limits = [
                lb for lb in linebreaks if sents[0].start < lb <= quote.end + 1
            ]
            if linebreak_limits:
                return [s for s in sents if s.end <= max(linebreak_limits)]
        return sents


def prep_text_for_quote_detection(t: str, fix_plural_possessives: bool = True) -> str:
    """
    Sorts out some common issues that trip up the quote detector.
    Works best one paragraph at a time -- use prep_document_for_quote_detection for the whole doc.

    - replaces consecutive apostrophes with a double quote (no idea why this happens but it does)
    - adds spaces before or after double quotes that don't have them
    - if enabled, fixes plural possessives by adding an "x", because the hanging apostrophe can trigger quote detection.
    - adds a double quote to the end of paragraphs that are continuations of quotes and thus traditionally don't end with quotation marks

    Input:
        t (str) - text to be prepped, preferably one paragraph
        fix_plural_possessives (bool) - enables fix_plural_possessives

    Output:
        t (str) - text prepped for quote detection
    """
    if not t:
        return

    t = t.replace("''", '"')
    if fix_plural_possessives:
        t = re.sub(r"(.{3,8}s\')(\s)", r"\1x\2", t)
    while re.search(constants.DOUBLE_QUOTES_NOSPACE_REGEX, p):
        match = re.search(constants.DOUBLE_QUOTES_NOSPACE_REGEX, p)
        if (
            len(re.findall(constants.ANY_DOUBLE_QUOTE_REGEX, p[: match.start()])) % 2
            != 0
        ):
            replacer = '" '
        else:
            replacer = ' "'
        p = p[: match.start()] + replacer + p[match.end() :]
    if (
        not (p[0] == "'" and p[-1] == "'")
        and p[0] in constants.ALL_QUOTES
        and len(re.findall(constants.ANY_DOUBLE_QUOTE_REGEX, p[1:])) % 2 == 0
    ):
        p += '"'
    return p.strip()


def prep_document_for_quote_detection(t: str, para_char: str = "\n") -> str:
    """
    Splits text into paragraphs (on para_char), runs prep_text_for_quote_detection on all paragraphs, then reassembles with para_char.

    Input:
        t (str) - document to prep for quote detection
        para_char (str) - paragraph boundary in t

    Output:
        document prepped for quote detection
    """
    return para_char.join(
        [prep_text_for_quote_detection(t) for t in t.split(para_char) if t]
    )



from collections import namedtuple
import regex as re
from itertools import zip_longest
from spacy.tokens import Doc, Span, Token
from typing import Iterable, Union, List, Literal
from .constants import (
    _reporting_verbs,
    _VERB_MODIFIER_DEPS,
    QUOTATION_MARK_PAIRS,
    ALL_QUOTES,
    BRACK_REGEX,
    DOUBLE_QUOTES,
    DOUBLE_QUOTES_NOSPACE_REGEX,
)
from spacy.symbols import VERB, PUNCT

DQTriple: tuple[list[Token], list[Token], Span] = namedtuple(
    "DQTriple", ["speaker", "cue", "content"]
)


def filter_cue_candidates(tok):
    return all([tok.pos == VERB, tok.lemma_ in _reporting_verbs])


def filter_speaker_candidates(ch, i, j):
    return all(
        [
            ch.pos != PUNCT,
            ((ch.i >= i and ch.i >= j) or (ch.i <= i and ch.i <= j)),
        ]
    )


def filter_quote_tokens(tok: Token, qtok_idx_pairs: List[tuple]) -> bool:
    return any(i <= tok.i <= j for i, j in qtok_idx_pairs)


def get_qtok_idx_pairs(doc: Union[Doc, Span]) -> List[tuple]:
    qtoks = [tok for tok in doc if tok.is_quote or (re.match(r"(\n)+", tok.text))]
    qtok_idx_pairs = [(-1, -1)]
    for n, q in enumerate(qtoks):
        if (
            not bool(q.whitespace_)
            and q.i not in [q_[1] for q_ in qtok_idx_pairs]
            and q.i > qtok_idx_pairs[-1][1]
        ):
            for q_ in qtoks[n + 1 :]:
                if (ord(q.text), ord(q_.text)) in QUOTATION_MARK_PAIRS:
                    qtok_idx_pairs.append((q.i, q_.i))
                    break
    return qtok_idx_pairs[1:]


def expand_noun(tok: Token) -> list[Token]:
    """Expand a noun token to include all associated conjunct and compound nouns."""
    tok_and_conjuncts = [tok] + list(tok.conjuncts)
    compounds = [
        child
        for tc in tok_and_conjuncts
        for child in tc.children
        if child.dep_ == "compound"
    ]
    return tok_and_conjuncts + compounds


def expand_verb(tok: Token) -> list[Token]:
    """Expand a verb token to include all associated auxiliary and negation tokens."""
    verb_modifiers = [
        child for child in tok.children if child.dep in _VERB_MODIFIER_DEPS
    ]
    return [tok] + verb_modifiers


def get_sent_idxs(span):
    indexes = [
        n
        for n, s in enumerate(span.doc.sents)
        if (s.start <= span.start <= s.end) or (s.start <= span.end <= s.end)
    ]
    return indexes[0], indexes[-1]


def line_break_window(span):
    """
    Finds the boundaries of the paragraph containing doc[i:j].
    """
    lb_tok_idxs = (
        [0]
        + [tok.i for tok in span.doc if re.match(r"\n", tok.text)]
        + [span.doc[-1].i]
    )
    for i_, j_ in zip_longest(lb_tok_idxs, lb_tok_idxs[1:]):
        if i_ <= span.start and j_ >= span.end:
            return (i_, j_)
    else:
        return (None, None)


def windower(span, method: Literal["overlap", "linebreaks"] = None):
    if method == "overlap":
        return [
            sent
            for sent in span.doc.sents
            if (sent.start < span.start < sent.end)
            or (sent.start < span.end < sent.end)
        ]
    else:
        i_sent, j_sent = get_sent_idxs(span)
        sents = (
            list(span.doc.sents)[i_sent - 1 : j_sent + 2]
            if i_sent > 0
            else list(span.doc.sents)[: j_sent + 2]
        )
        if method == "linebreaks":
            linebreaks = (
                [0]
                + [tok.i for tok in span.doc if re.match(r"\n", tok.text)]
                + [span.doc[-1].i]
            )
            linebreak_limits = [
                lb for lb in linebreaks if sents[0].start < lb <= span.end + 1
            ]
            if linebreak_limits:
                return [s for s in sents if s.end <= max(linebreak_limits)]
        return sents


def old_windower(span, lb_boundaries=False) -> Iterable:
    if lb_boundaries:
        i_, j_ = line_break_window(span)
        if i_ is not None and j_ is not None:
            return list(span.doc[i_ + 1 : j_ - 1].sents)
        else:
            return [span]

    else:
        i, j = span.start, span.end
        return [
            sent
            for sent in span.doc.sents
            # these boundary cases are a subtle bit of work...
            if (
                (sent.start < i and sent.end >= i - 1)
                or (sent.start <= j + 1 and sent.end > j)
            )
        ]


def para_quote_fixer(p, exp: bool = False):
    if not p:
        return
    p = p.strip()
    p = p.replace("''", '"')
    p = re.sub(r"(.{3,8}s\')(\s)", r"\1x\2", p)

    while re.search(DOUBLE_QUOTES_NOSPACE_REGEX, p):
        match = re.search(DOUBLE_QUOTES_NOSPACE_REGEX, p)
        if (
            len(re.findall(BRACK_REGEX.format(DOUBLE_QUOTES), p[: match.start()])) % 2
            != 0
        ):
            replacer = '" '
        else:
            replacer = ' "'
        p = p[: match.start()] + replacer + p[match.end() :]
    if (
        not (p[0] == "'" and p[-1] == "'")
        and p[0] in ALL_QUOTES
        and len(re.findall(BRACK_REGEX.format(DOUBLE_QUOTES), p[1:])) % 2 == 0
    ):
        p += '"'
    return p


def prep_text_for_quote_detection(t, para_char="\n", exp: bool = False):
    return para_char.join(
        [para_quote_fixer(p, exp=exp) for p in t.split(para_char) if p]
    )



"""
Rewritten with less overhead.
"""
import spacy
import numpy as np
from .quotes import direct_quotations
from . import constants
from . import attribution_helpers

class Attributor:
    """
    TODO: Add manual speaker matching
    """

    def __init__(
        self,
        text: str = None,
        coref_nlp: str = "en_coreference_web_trf",
        base_nlp: str = "en_core_web_lg",
        prune: bool = True
    ):
        self.coref_nlp = spacy.load(coref_nlp)
        self.base_nlp = spacy.load(base_nlp)
        self.prune = prune
        if text:
            self.attribute(text)

    def expand_match(self, match=None):
        if not match:
            match = self.quote_matches
            
        if isinstance(match, list):
            for m in match:
                self.expand_match(m)
        else:
            for m_ in ["quote", "cluster", "person"]:
                if getattr(match, f"{m_}_index", None) is not None:
                    i = eval(f"self.{m_}s")
                    v = getattr(match, f"{m_}_index")
                    data = (
                        attribution_helpers.format_cluster(i[v])
                        if m_ == "cluster"
                        else i[v]
                    )
                    print(m_.upper(), f": {v}" "\n", data, "\n")

    def attribute(self, t: str):
        """
        Top level function. Parses text, matches quotes to clusters and gets ent matches.
        Input:
            t (str) - text file to be analyzed and attributed

        Output:
            self.quote_matches (list[QuoteClusterMatch]) - list of quotes
        """
        self.parse_text(t)
        self.quote_matches = self.get_matches()

    def parse_text(self, t: str):
        """
        Imports text, gets coref clusters, copies coref clusters, finds PERSONS and gets NER matches.

        Input:
            t (string) - formatted text of an article

        Ouput:
            self.coref_doc - spacy coref-parsed doc
            self.doc - spacy doc with coref clusters
            self.clusters - coref clusters
            self.quotes - list of textacy-extracted quotes
            self.persons - list of PERSON entities
        """
        # instantiate spacy doc
        self.coref_doc = self.coref_nlp(t)
        self.doc = self.base_nlp(t)

        # extract quotations
        self.quotes = [q for q in direct_quotations(self.doc)]

        # extract coref clusters and clone to doc
        self.clusters = [
            attribution_helpers.clone_cluster(cluster, self.doc)
            for k, cluster in self.coref_doc.spans.items()
            if k.startswith("coref")
        ]
        if self.prune:
            self.clusters = [attribution_helpers.prune_cluster_people(cluster)
                for cluster in self.clusters
            ]

        self.persons = [e for e in self.doc.ents if e.label_ == "PERSON"]
        return

    def get_matches(self):
        pairs_dicto = self.make_pairs()
        arrays = {k: self.make_matrix(k, v) for k, v in pairs_dicto.items()}

        big_matrix = np.concatenate(
            (
                np.transpose(
                np.nonzero(
                    arrays["quotes_persons"].dot(
                        arrays["clusters_persons"].T
                    )
                )
            ),
            np.transpose(np.nonzero(arrays["quotes_clusters"])),
            )
        )

        results = sorted(list(set(
            [constants.QuoteClusterMatch(i,j) for i,j in big_matrix]
            )), key=lambda m: m.quote_index)

        return results

    def make_pairs(self) -> dict:
        """
        Creates quote/person, quote/cluster and cluster/person pairs for resolution and cleaning.

        TODO: Ensure pronouns aren't being skipped!
        TODO: Make ratio threshold a variable
        """
        pairs_dicto = {
            p: []
            for p in [
                "quotes_persons",
                "quotes_clusters",
            ]
        }

        for quote_index, quote in enumerate(self.quotes):
            pairs_dicto["quotes_clusters"] += [
                (quote_index, cluster_index)
                for cluster_index, cluster in enumerate(self.clusters)
                for span in cluster
                if attribution_helpers.compare_quote_to_cluster_member(quote, span)
            ]

            pairs_dicto["quotes_persons"] += [
                (quote_index, person_index)
                for person_index, person in enumerate(self.persons)
                if attribution_helpers.span_contains(quote, person)
            ]

            pairs_dicto["quotes_clusters"] += [
                (quote_index, cluster_index)
                for cluster_index, cluster in enumerate(self.clusters)
                if quote_index
                not in [m[0] for m in set(pairs_dicto["quotes_clusters"])]
                if quote.speaker[0].text in [p.text for p in self.persons]
                if attribution_helpers.get_manual_speaker_cluster(quote, cluster)
            ]

        pairs_dicto["clusters_persons"] = [
            (cluster_index, person_index)
            for person_index, person in enumerate(self.persons)
            for cluster_index, cluster in enumerate(self.clusters)
            for span in cluster
            if attribution_helpers.span_contains(person, span)
            if not attribution_helpers.pronoun_check(span)
        ]
        return pairs_dicto

    def make_matrix(self, key, pairs):
        x, y = key.split("_")
        m = np.zeros([len(self.__getattribute__(_)) for _ in [x, y]])
        for i, j in pairs:
            m[i, j] = 1
        return m
    
    def print_clusters(self):
        """
        Print clusters with duplicate text removed. For easier analysis!
        """
        for n, cluster in enumerate(self.clusters):
            print(n, set(t.text for t in cluster))



from spacy.symbols import aux, auxpass, csubj, dobj, neg, nsubj
from collections import namedtuple
from spacy.tokens import Span, Token

DQTriple: tuple[list[Token], list[Token], Span] = namedtuple(
    "DQTriple", ["speaker", "cue", "content"]
)

QuoteClusterMatch: tuple[int, int] = namedtuple(
    "QuoteClusterMatch", ["quote_index", "cluster_index"], defaults=(None, None)
)

Boundaries: tuple[int, int] = namedtuple("boundaries", ["start", "end"])

"""
Constants for token/entity matching
"""
MIN_SPEAKER_DIFF = 5
MIN_ENTITY_DIFF = 2
MIN_QUOTE_LENGTH = 3

"""
Constants for textacy quote identification
"""
_ACTIVE_SUBJ_DEPS = {csubj, nsubj}
_VERB_MODIFIER_DEPS = {aux, auxpass, neg}

"""
For prepping text for quote detection.
"""
ALL_QUOTES = "‹「`»」‘\"„›”‚’'』『«“"
DOUBLE_QUOTES = '‹「」»"„『”‚』›«“'
BRACK_REGEX = r"[{}]"
DOUBLE_QUOTES_NOSPACE_REGEX = r"(?<=\S)([{}])(?=\S)".format(DOUBLE_QUOTES)

"""
Ordinal points of the token.is_quote characters, matched up by start and end.

source:
switch = "\"\'"
start = "“‘```“‘«‹「『„‚"
end = "”’’’’”’»›」』”’"

"""
QUOTATION_MARK_PAIRS = {
    (34, 34),  # " "
    (39, 39),  # ' '
    (96, 8217),
    (171, 187),
    (8216, 8217),
    (8218, 8217),
    (8220, 8221),
    (8222, 8221),
    (8249, 8250),
    (12300, 12301),
    (12302, 12303),
    (8220, 34),
    (8216, 34),
    (96, 34),
    (171, 34),
    (8249, 34),
    (12300, 34),
    (12302, 34),
    (8222, 34),
    (8218, 34),
    (34, 8221),
    (34, 8217),  # " ’
    (34, 10),
    (39, 10),
    (96, 10),
    (171, 10),
    (8216, 10),
    (8218, 10),
    (8249, 10),
}

_reporting_verbs = {
    "according",
    "accuse",
    "acknowledge",
    "add",
    "admit",
    "agree",
    "allege",
    "announce",
    "argue",
    "ask",
    "assert",
    "believe",
    "blame",
    "charge",
    "cite",
    "claim",
    "complain",
    "concede",
    "conclude",
    "confirm",
    "contend",
    "continue",
    "criticize",
    "declare",
    "decline",
    "deny",
    "describe",
    "disagree",
    "disclose",
    "estimate",
    "explain",
    "fear",
    "hope",
    "insist",
    "maintain",
    "mention",
    "note",
    "observe",
    "order",
    "post",
    "predict",
    "promise",
    "read",
    "recall",
    "recommend",
    "reply",
    "report",
    "say",
    "scream",
    "state",
    "stress",
    "suggest",
    "tell",
    "testify",
    "think",
    "tweet",
    "urge",
    "warn",
    "worry",
    "write",
}



from .constants import (
    MIN_ENTITY_DIFF,
    MIN_SPEAKER_DIFF,
    Boundaries,
    QuoteClusterMatch,
)
from .quote_helpers import DQTriple
from spacy.tokens import Span, SpanGroup, Token, Doc
from typing import Union, Literal, Tuple, Iterable
import statistics
from rapidfuzz import fuzz
import numpy as np


def get_cluster_people_scores(
    cluster: SpanGroup, scorer: Literal["prat", "cos"] = "prat"
) -> Tuple[list, float]:
    """
    Calculates average similarity between any two PERSONS in the cluster.
    These scores are used to exclude "odd man out" cluster members.

    TODO: tweak cutoff value, or at least make it flexible

    Input:
        cluster (SpanGroup) - coref cluster
        scorer (str) - what score to use to determine similarity. can be 'prat' (partial ratio) or 'cos' (cosine similarity).

    Output:
        list(tuple) - index, span, average score for each span in the cluster
        cutoff (float) - minimum score for keeping cluster member (mean - 2stdev)
    """
    if scorer == "prat":
        score_func = lambda s1, s2: fuzz.partial_ratio(s1.text, s2.text)
    elif scorer == "cos":
        score_func = lambda s1, s2: s1.similarity(s2)

    # filter out non-persons
    cluster_ = [span for span in cluster if person_check(span)]

    all_scores = []
    for n, span in enumerate(cluster_):
        scores = [score_func(span, span_) for span_ in cluster_]
        if scores:
            all_scores.append((n, span, sum(scores) / len(scores)))
    if len(all_scores) > 1:
        cutoff = statistics.mean([c[-1] for c in all_scores]) - 2 * statistics.stdev(
            [c[-1] for c in all_scores]
        )
        return sorted(all_scores, key=lambda k: k[-1]), cutoff
    else:
        return [], None


def prune_cluster_people(cluster: SpanGroup, scorer="prat") -> list:
    """
    Removes outlier PERSONS from a cluster, based on provided score.
    TODO: SpanGroup instead of list?

    Input:
        cluster (SpanGroup) - a coref cluster

    Output:
        list - coref cluster with outlier PERSONS removed
    """
    scores, cutoff = get_cluster_people_scores(cluster, scorer=scorer)
    filtered = [c[1] for c in scores if c[-1] < cutoff]
    return [c for c in cluster if c not in filtered]


def clone_cluster(cluster: SpanGroup, destination_doc: Doc):
    """
    For copying a coref cluster to a different Doc.

    Necessary because I'm using multiple models (of different sizes) to do coreferencing and other NLP tasks.
    It's easier to consolidate the clusters than to combine the tasks into one model.
    """
    return SpanGroup(
        doc=destination_doc,
        spans=([destination_doc[span.start : span.end] for span in cluster]),
    )


def filter_duplicate_ents(ents) -> tuple:
    """
    Removes duplicate entities by text.
    """
    ent_bucket = []

    for e in ents:
        if e.text in [e.text for e in ent_bucket]:
            continue
        else:
            ent_bucket.append(e)

    return ent_bucket


def get_boundaries(t: Union[Token, Span, list, DQTriple]) -> Boundaries:
    """
    Convenience function that returns a Boundaries tuple with the start and the end character of t.

    Necessary because Token and Span have differently named attributes.

    Input:
        t (Token or Span) - spacy token or span object

    Output:
        Boundares(start, end) with start character and end character of t.
    """
    if isinstance(t, Token):
        return Boundaries(t.idx, t.idx + len(t))

    elif isinstance(t, Span):
        return Boundaries(t.start_char, t.end_char)

    elif isinstance(t, tuple):  # isinstance won't recognize DQTriple
        return Boundaries(
            get_boundaries(t.speaker[0]).start,
            get_boundaries(t.speaker[-1]).end,
        )

    else:
        raise TypeError("input needs to be Token, Span or DQTriple!")


def get_text(t: Union[Token, Span, list]) -> str:
    """
    Convenience function, because quote speakers are lists of tokens.
    """
    if isinstance(t, Token) or isinstance(t, Span):
        return t.text
    if isinstance(t, list):
        return " ".join([t_.text for t_ in t])


def span_contains(
    t1: Union[Span, Token, DQTriple], t2: Union[Span, Token, DQTriple]
) -> bool:
    """
    Does t1 contain t2 or v/v? Assumes both are from the same doc, or at least same index.

    Uses get_boundaries beacuse quote speakers are tokens but entities are spans.

    TODO: verify same doc

    Input:
        t1 and t2 - spacy spans or tokens

    Output:
        bool - whether either t1 contains the other t2

    """
    b1, b2 = tuple([get_boundaries(t) for t in [t1, t2]])

    if (b1.start <= b2.start and b1.end >= b2.end) or (
        b2.start <= b1.start and b2.end >= b1.end
    ):
        return True

    else:
        return False


def format_cluster(cluster):
    return list(set([c.text for c in cluster if c[0].pos_ != "PRON"]))


def pronoun_check(t: Span, doc: Doc = None):
    """
    Checks to see if t is a single pronoun.
    """
    if len(t) == 1:
        if doc:
            return doc[t[0].i].pos_ == "PRON"
        else:
            return t[0].pos_ == "PRON"
    return False


def person_check(span: Span):
    """
    Convenience function.
    """
    try:
        return span.ents[0].label_ == "PERSON"
    except IndexError:
        return False


def get_manual_speaker_cluster(quote, cluster):
    """
    If the match doesn't have a cluster, find any speakers in clusters that match manually.

    The idea here is if the speaker is "Rosenberg" to pull "Detective Jeff Rosenberg" from the clusters.

    TODO: Add this back into attribution code, and a test.
    """
    if any([len(quote.speaker) > 1, quote.speaker[0].pos_ != "PRON"]):
        speaker = " ".join([s.text for s in quote.speaker])
        for span in cluster:
            if speaker in span.text:
                return True
    else:
        return False


def compare_quote_to_cluster(
    quote: DQTriple,
    cluster: SpanGroup,
):
    """
    Finds first span in cluster that matches (according to compare_quote_to_cluster_member) with provided quote.

    TODO: Doesn't consider further matches. Is this a problem?

    Input:
        quote - textacy quote object
        cluster - coref cluster

    Output:
        cluster_index (int) - index of cluster member that matches quote (or -1, if none match)
    """
    try:
        return next(
            cluster_index
            for cluster_index, cluster_member in enumerate(cluster)
            if compare_quote_to_cluster_member(quote, cluster_member)
        )
    except StopIteration:
        return -1


def compare_quote_to_cluster_member(quote: DQTriple, span: Span):
    """
    Compares the starting character of the quote speaker and the cluster member as well as the quote speaker sentence and the cluster member sentence to determine equivalence.

    Input:
        q (quote triple) - one textacy quote triple
        cluster_member - one spacy-parsed entity cluster member

    Output:
        bool
    """
    # filters out very short strings
    if span[0].pos_ != "PRON" and len(span) < 2 and len(span[0]) < 4:
        return False
    if abs(quote.speaker[0].sent.start_char - span.sent.start_char) < MIN_SPEAKER_DIFF:
        if abs(quote.speaker[0].idx - span.start_char) < MIN_SPEAKER_DIFF:
            return True
    if span.start_char <= quote.speaker[0].idx:
        if span.end_char >= (quote.speaker[-1].idx + len(quote.speaker[-1])):
            return True
    return False


def compare_spans(
    s1: Span,
    s2: Span,
) -> bool:
    """
    Compares two spans to see if their starts and ends are less than min_entity_diff.

    If compare

    Input:
        s1 and s2 - spacy spans
        min_entity_diff (int) - threshold for difference in start and ends
    Output:
        bool - whether the two spans start and end close enough to each other to be "equivalent"

    """
    return all(
        [
            abs(getattr(s1, attr) - getattr(s2, attr)) < MIN_ENTITY_DIFF
            for attr in ["start", "end"]
        ]
    )


def quick_ent_analyzer(
    quote_person_pairs,
    quote_cluster_pairs,
    quote_ent_pairs,
    cluster_ent_pairs,
    cluster_person_pairs,
    person_ent_pairs,
):
    def attributed_quote_filter(idx, ent_matches):
        return idx in [m.quote_index for m in ent_matches]

    ent_matches = []
    quote_matches = []

    for qe in quote_ent_pairs:
        ent_matches.append(QuoteEntMatch(qe[0], None, None, qe[1]))

    for qp in quote_person_pairs:
        for pe in person_ent_pairs:
            if qp[1] == pe[0] and not attributed_quote_filter(qp[0], ent_matches):
                ent_matches.append(QuoteEntMatch(qp[0], None, pe[0], pe[1]))

    for qc in quote_cluster_pairs:
        if qc[1] == None and not attributed_quote_filter(qc[0], ent_matches):
            ent_matches.append(QuoteEntMatch(qc[0]))
        else:
            quote_matches.append(QuoteClusterMatch(qc[0], qc[1]))
            for ce in cluster_ent_pairs:
                if qc[1] == ce[0] and not attributed_quote_filter(qc[0], ent_matches):
                    ent_matches.append(QuoteEntMatch(qc[0], ce[0], None, ce[1]))

            for cp in cluster_person_pairs:
                if qc[1] == cp[0]:
                    for pe in person_ent_pairs:
                        if cp[1] == pe[0] and not attributed_quote_filter(
                            qc[0], ent_matches
                        ):
                            ent_matches.append(
                                QuoteEntMatch(qc[0], cp[0], pe[0], pe[1])
                            )

    return sorted(list(set(ent_matches)), key=lambda m: m.quote_index)
